#ifndef AETH_ARPHANDLER_H_
#define AETH_ARPHANDLER_H_

#include <alib.h> // IWYU pragma: keep

#include <alib_vector.h>
#include <alib_circularbufferraw.h>

#include <anet_ipaddress.h>

#include <aeth_ethernetaddress.h>
#include <aeth_frame_eth.h>

namespace HIPASE
{

class CAeth_EthHost;

class CAeth_ArpHandler
{
protected:
    struct CRequestState;

    inline CAeth_ArpHandler(CRequestState *pArray, unsigned size, unsigned reqtimeout = 3, unsigned stabletimeout = 10000);

    /* Wieviele Arp-Anfragen gesendet werden sollen bevor die IP Addresse als nicht erreichbar eingetragen wird */
    typedef std::integral_constant<unsigned char, 4> arpretries_t;

public:
    enum EArpRequestType
    {
        eArpRequestType_NoRequest,
        eArpRequestType_Request,
        eArpRequestType_Probe,
        eArpRequestType_Announce
    };
#if 0
    CAeth_ArpHandler(void *pArray, unsigned size, unsigned reqtimeout = 500, unsigned stabletimeout = 5000)
    : m_ArpTable(reinterpret_cast<CRequestState *>(pArray), size / sizeof(CRequestState)),
      m_RequestTimeout(reqtimeout),
      m_StableTimeout(stabletimeout)
    {}
#endif

    /** \brief Finde Mac Addresse zu IP
     *
     * Sucht in der Tabelle eine Mac Adresse zur angegebenen IP.
     * Der zurückgegebene Pointer zeigt auf die interne Struktur,
     * und kann somit ungültig werden. Er sollte nicht nach weiteren Aufrufen
     * des Handlers benutzt werden.
     *
     * @param ipaddr
     * @return Einen Pointer auf die Mac Adresse oder nullptr
     */
    const CAeth_EthernetAddress *find(CAnet_AddressIp4 ipaddr) const;

    /** \breif Finde Mac Adresse zu IP und bereite ARP Request vor falls notwendig
     *
     * Sucht in der Tabelle eine Mac Adresse zur angegebenen IP. Bei Misserfolg
     * wird zugleich der ARP-Request vorbereitet.
     *
     * Der zurückgegebene Pointer zeigt auf die interne Struktur,
     * und kann somit ungültig werden. Er sollte nicht nach weiteren Aufrufen
     * des Handlers benutzt werden.
     *
     * Die Rückgabe-Werte sind so zu behandeln:
     * \e return != \e nullptr : aktiver Eintrag gefunden, \e sendImmediate \e ist invalid
     * \e return == \e nullptr, sendImmediate war \b true, ist nach Aufruf immer noch \b true : ein ARP-Request ist nötig, der Aufrufer sollte diesen sobald möglich verschicken
     *
     * @param ipaddr
     * @param timestamp
     * @param[in, out] Wen Aufrufer true übergibt, muß dieser den eventuellen ARP Request verschicken
     * @return Einen Pointer auf die Mac Adresse oder nullptr
     */
    const CAeth_EthernetAddress *findImmediate(CAnet_AddressIp4 ipaddr, unsigned timestamp, bool &sendImmediate);


    FORCEINLINE const CAeth_EthernetAddress *find(uint32_t ipaddr) const
    {
        return find(CAnet_AddressIp4::fromIp(ipaddr));
    }

    /** \brief Erzeuge ARP Request
     *
     * Requests können nur in bestimmten Zuständen generiert werden, um zu verhindern das mehrere
     * ausständige Requests unterwegs sind
     *
     * @param timestamp
     * @param ipaddr
     * @return true wenn nach der Funktion der IP-Eintrag im Request Zustand ist
     */
    bool queueRequest(CAnet_AddressIp4 ipaddr, unsigned timestamp, bool force = false);


    FORCEINLINE static constexpr CAnet_AddressIp4 getArpSourceIp(const CAnet_AddressIp4 hostIp, CAnet_AddressIp4 ipAddress, EArpRequestType type = eArpRequestType_Request)
    {
        return type == eArpRequestType_Announce ? ipAddress : (type == eArpRequestType_Probe ? CAnet_AddressIp4() : hostIp);
    }

    /** \brief generiere eine Arp Request Frame.
     *
     * @param host
     * @param rawBuffer
     * @param ipAddress
     */
    static void *prepareArpRequest(void *rawBuffer, const CAeth_EthernetAddress &hostMac, const CAnet_AddressIp4 hostIp, CAnet_AddressIp4 ipAddress, EArpRequestType type = eArpRequestType_Request, bool byteswap = false)
    {
#if 0
        /* Standard ARP Request für IP4 */
        return CAeth_EthernetArpFrame::writeFullIP4Arp(rawBuffer, CAeth_EthernetArpFrame::eOperation_Request,
                hostMac, getArpSourceIp(hostIp, ipAddress, type),
                CAeth_EthernetAddress(), ipAddress,
                byteswap);
#endif
    }

    static unsigned prepareArpResponse(const CAeth_EthHost &host, void *rawBuffer, const CAeth_EthernetAddress &tmac, CAnet_AddressIp4 ipAddress);

    FORCEINLINE static void *prepareArpProbe(void *rawBuffer, const CAeth_EthernetAddress &hostMac, const CAnet_AddressIp4 hostIp, CAnet_AddressIp4 ipAddress, bool byteswap = false)
    {
        return prepareArpRequest(rawBuffer, hostMac, hostIp, ipAddress, eArpRequestType_Probe, byteswap);
    }
    FORCEINLINE static void *prepareArpAnnounce(void *rawBuffer, const CAeth_EthernetAddress &hostMac, const CAnet_AddressIp4 hostIp, CAnet_AddressIp4 ipAddress, bool byteswap = false)
    {
        return prepareArpRequest(rawBuffer, hostMac, hostIp, ipAddress, eArpRequestType_Announce, byteswap);
    }

    /** \brief Behandle Timeouts und andere automatische Zustandsänderungen
     *
     * TODO: das inkrementell machen
     * @param timestamp
     */
//    unsigned generatePackets(const CAeth_EthHost &host, void **buffers, unsigned maxNum, unsigned maxsteps);

    void updateTable(unsigned tick, unsigned maxnum = ~0U);


    /** \brief Halte einen Eintrag am Leben indem der Timeout reset wird
     *
     * Bei bestehenden Verbindungen können so zusätzliche, periodische ARP Requests vermieden werden.
     * Es wird jedoch niemals eine neue Mac eingetragen, im Zweifelsfall wird der Aufruf einen
     * Eintrag als veraltet markieren und ein ARP Request wird eventuelle Änderungen anfordern
     *
     * @param ipAddress
     * @param macAddress
     * @param ts
     * @return true, wenn ein gültiger Eintrag nur aufgefrischt wurde.
     */
    bool refreshMapping(CAnet_AddressIp4 ipAddress, const CAeth_EthernetAddress &macAddress, unsigned ts);

    /** \brief Verarbeite empfangene Arp Pakete
     *
     * Tested Paket auf Gültigkeit und aktualisiert die Tabelle wenn notwendig.
     * Falls eine Antwort notwendig ist, wird die frame in-place modifiziert und
     * der Aufrufer muß das Paket an den Sender zurückschicken.
     *
     * @param host
     * @param ethFrame
     * @return
     */
    EArpRequestType processIp4ArpPackets(const CAeth_EthHost &host, unsigned timestamp, const CAeth_EthernetArpFrame::CArpPacket4 &ethFrame, unsigned short operation);

    template <typename TCallback>
    bool sendRequest(TCallback &&cb)
    {
        if (!m_QueuedRequest.empty() && cb(m_QueuedRequest.front()))
        {
            m_QueuedRequest.free_front();
            return true;
        }
        return false;
    }

private:
    FORCEINLINE static constexpr uint32_t convert32(uint32_t val)
    {
        return Alib_BitUtils::convertBE(val);
    }
    FORCEINLINE static constexpr uint16_t convert16(uint16_t val)
    {
        return Alib_BitUtils::convertBE(val);
    }

    FORCEINLINE const CRequestState *findInternal(CAnet_AddressIp4 ipaddr) const;

    FORCEINLINE const CRequestState *updateEntry(CAnet_AddressIp4 ipAddress, const CAeth_EthernetAddress &macAddress, unsigned ts, bool checkstate = true);

    CAlib_VectorHead<CRequestState> m_ArpTable;
    CAlib_CircularBufferRaw<CAnet_AddressIp4, 16> m_QueuedRequest;
    unsigned m_RequestTimeout;
    unsigned m_StableTimeout;
    unsigned m_LastOffset;
};



}  // namespace HIPASE

#endif /* AETH_ARPHANDLER_H_ */
