
#ifndef ALIB_CIRCULARBUFFERRWC_H_
#define ALIB_CIRCULARBUFFERRWC_H_

#include <alib.h>
#include <fws_asan.h>

#include <cstring>
#include <utility>
#include <iterator>
#include <type_traits>

#include <alib_atomic.h>
#include <alib_circularbufferiterator.h>

namespace HIPASE
{

enum EAlib_CircularBufferType
{
    eAlib_CircularBufferType_OneEmpty,
    eAlib_CircularBufferType_Fillcount,
    eAlib_CircularBufferType_Mirroring,
    eAlib_CircularBufferType_RwCounts,
    eAlib_CircularBufferType_Indices,
    eAlib_CircularBufferType_LastFlag,
};

#if 0
/* benötigt libcstd++ von gcc 5.1, unmöglich dies sauber zu testen */
template <typename TElement, std::size_t N, bool = std::is_trivially_default_constructible<TElement>::value>
class CAlib_CircularBufferRwc;
#else
template <typename TElement, std::size_t N>
class CAlib_CircularBufferRwc;
#endif

namespace Alib_CircularBuffer_Internal
{

/** \brief Verwaltet einen Puffer fixer Größe, der als synchrone nichtblockende Queue zwischen 2 Threads benutzt werden kann
 *
 * Die Vorraussetzung ist, das nur jeweils ein thread liest und schreibt. Damit ist sichergestellt das eine der beiden Variablen
 * fix bleibt, m_Front wird nur vom Empfänger inkrementiert und m_Back nur vom Sender.
 * Die Synchronisation passiert ohne locks ("lock free").
 *
 * Der Empfänger muß zuerst überprüfen, ob k >= 1 elemente vorhanden sind,
 * danach könnte der Sender ein oder mehrere Elemente einfügen bevor die Operation abgeschlossen ist,
 * jedoch sind immer noch mindestens k Elmente vorhanden.
 *
 * Der Sender muß zuerst testen, ob noch k >= 1 Plätze frei sind,
 * danach könnte der Empfänger noch Plätze freigeben bevor die operation abgeschlossen ist,
 * jedoch ist immer noch mindestens k Plätze frei.
 *
 * Die gewählte Implementierung (http://en.wikipedia.org/wiki/Circular_buffer#Read_.2F_Write_Counts) benutzt 2 Variablen zum synchronisieren
 * der Zugriffe (= feststellen ob Puffer voll/leer ist).
 * Exceptions wurden, wo möglich, vermieden um diese Klasse auch in interrupt-routinen verwenden zu können.
 * Stattdessen signalisieren pop und push Fehler durch Rückgabewerte.
 *
 * Die Implementierung sieht nur Verwendung mit trivialen Datentypen vor, es wird wirklich nur der notwendige Speicher verwaltet,
 * Instanzen werden nicht initialisiert oder dealloziert. Dies hat den Vorteil, dass die Anforderungen an den Datentyp minimal sind,
 * ansonsten gibts Probleme mit eg. volatile Typen.
 *
 *
 * Empfangs-Operation:
 * #clear, #pop_front #pop_front_n, #pop_front(reference)
 *
 * Sende-Operation:
 * #alloc_back, #alloc_back_n(size_type), #alloc_back_n(size_type&&, bool), #push_back
 *
 * \tparam TElement
 */
template <typename TElement>
class CAlib_CircularBufferRwcHead
{
    static constexpr FWS_ASAN_SANTYPE(TElement) getSanContext()
    {
        return {};
    }
public:
	typedef TElement value_type;
    typedef TElement* pointer;
    typedef const TElement *const_pointer;
	typedef TElement& reference;
	typedef const TElement& const_reference;
	typedef typename CAlib_CircularBufferIterator<TElement>::difference_type difference_type;
	typedef std::size_t size_type;

    typedef CAlib_CircularBufferIterator<TElement> iterator;
    typedef CAlib_CircularBufferIterator<const TElement> const_iterator;
    typedef std::reverse_iterator<const_iterator>  const_reverse_iterator;
    typedef std::reverse_iterator<iterator>        reverse_iterator;

	typedef std::integral_constant<EAlib_CircularBufferType, eAlib_CircularBufferType_RwCounts> ringbufferimpl_t;
	typedef std::true_type readerwriterconcurrent_t;

    template <unsigned N>
    struct storage_t
    {
        typedef typename std::aligned_storage<sizeof(TElement) * (N > 0 ? N : 1), std::alignment_of<TElement>::value>::type type;
    };

    template <unsigned N>
    using static_type = CAlib_CircularBufferRwc<TElement, N>;

    CAlib_CircularBufferRwcHead &head()
    {
        return *this;
    }
    constexpr const CAlib_CircularBufferRwcHead &head() const
    {
        return *this;
    }

    constexpr CAlib_CircularBufferRwcHead() noexcept
            : m_pArray(), m_NumElements(), m_PopCount(), m_Front(), m_PushCount(), m_Back()
    {}

protected:




#endif
protected:
	/** \brief Pointer zu den Daten */
	pointer m_pArray;
	/** \brief Anzahl der elemente im Array */
    std::size_t m_NumElements;
	/** \brief Anzahl der entfernten Elemente */
	CAlib_Atomic<std::size_t> m_PopCount;
	/** \brief Zeiger auf das erste Element */
	pointer m_Front;
	/** \brief Anzahl der hinzugefügten Elemente */
	CAlib_Atomic<std::size_t> m_PushCount;
	/** \brief Zeiger auf den ersten freien Platz */
	pointer m_Back;
};

} // namespace Alib_CircularBuffer_Internal

/** \brief Spezialisierung für Pointer
 *
 * Um auch zB. const int * ptr zu unterstützen, werden C-Style Casts benutzt
 */
template <typename TElement>
class CAlib_CircularBufferRwcHead : public Alib_CircularBuffer_Internal::CAlib_CircularBufferRwcHead<TElement>
{
	typedef Alib_CircularBuffer_Internal::CAlib_CircularBufferRwcHead<TElement> super_t;
public:

	constexpr CAlib_CircularBufferRwcHead()
    : super_t()
      {}

	/** \brief Erstellt einen neuen Ringbuffer über einen existierenden Array
	 *
	 * @param base Pointer zu dem Array
	 * @param size Größe des Arrays in elementen
	 */
	constexpr CAlib_CircularBufferRwcHead(typename super_t::pointer base, typename super_t::size_type size) noexcept
	: super_t(base, size)
	  {}

	/** \brief Erstellt einen neuen Ringbuffer über einen existierenden Array
	 *
	 * @param a Array
	 */
	template <std::size_t N>
	constexpr CAlib_CircularBufferRwcHead(TElement (&a)[N]) noexcept
	: super_t(a)
	  {}

    CAlib_CircularBufferRwcHead(CAlib_CircularBufferRwcHead &&other) noexcept
      : super_t(std::move(other))
    { }

    CAlib_CircularBufferRwcHead &operator=(CAlib_CircularBufferRwcHead &&other) noexcept
    {
        super_t::operator =(std::move(other));
        return *this;
    }

    /* Kopieren nicht erlaubt */
    CAlib_CircularBufferRwcHead(const CAlib_CircularBufferRwcHead &) = delete;
    CAlib_CircularBufferRwcHead &operator =(const CAlib_CircularBufferRwcHead &) = delete;
};




#if 0
template <typename TElement, std::size_t N, bool>
class CAlib_CircularBufferRwc : public CAlib_CircularBufferRwcHead<TElement>
{
public:
    typedef CAlib_CircularBufferRwcHead<TElement> head_t;
    FORCEINLINE constexpr CAlib_CircularBufferRwc()
	: CAlib_CircularBufferRwcHead<TElement>(m_Storage, N)
	  {}

private:
    TElement m_Storage[N];
};

template <typename TElement, std::size_t N>
class CAlib_CircularBufferRwc<TElement, N, false> : public CAlib_CircularBufferRwcHead<TElement>
{
public:
    typedef CAlib_CircularBufferRwcHead<TElement> head_t;
    FORCEINLINE CAlib_CircularBufferRwc()
    : CAlib_CircularBufferRwcHead<TElement>(reinterpret_cast<typename CAlib_CircularBufferRwcHead<TElement>::pointer>(&m_Storage), N)
      {}

private:
    typename CAlib_CircularBufferRwcHead<TElement>::template storage_t<N>::type m_Storage;
};
#else
template <typename TElement, std::size_t N>
class CAlib_CircularBufferRwc : public CAlib_CircularBufferRwcHead<TElement>
{
public:
    typedef CAlib_CircularBufferRwcHead<TElement> head_t;
    FORCEINLINE CAlib_CircularBufferRwc()
    : CAlib_CircularBufferRwcHead<TElement>(reinterpret_cast<typename CAlib_CircularBufferRwcHead<TElement>::pointer>(&m_Storage), N)
      {}

private:
    typename CAlib_CircularBufferRwcHead<TElement>::template storage_t<N>::type m_Storage;
};
#endif

} /* namespace HIPASE */
#endif /* ALIB_CIRCULARBUFFERRWC_H_ */
